#=======================================================================
#	$Id: Memory.pm,v 1.2 2005/09/07 10:07:17 pythontech Exp $
#	In-memory triple store
#	Copyright (C) 2004  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples::Memory;
use strict;
use PythonTech::Triples;

our @ISA = qw(PythonTech::Triples);

#-----------------------------------------------------------------------
#	With the object hash, each key is a resource URI (or literal);
#	the corresponding value is a 2-element list.
#	The first list element is a hash with key being predicate and
#	value being hash obj => 1.
#	The second list element is a hash with key being predicate and
#	value being hash subj => 1 (i.e. reverse mapping).
#	  res => [{pred => {obj => 1}}, {rpred => {subj=>1}}]
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {
	# See above
    };
    bless $self, $class;
    return $self;
}

sub add {
    my($self, $subj,$pred,$obj) = @_;
    unless ($self->{$subj}->[0]->{$pred}->{$obj}) {
	$self->{$subj}->[0]->{$pred}->{$obj} = 1;
	$self->{$obj}->[1]->{$pred}->{$subj} = 1;
    }
}

sub remove {
    my($self, $subj,$pred,$obj) = @_;
    if (exists $self->{$subj}
	&& exists $self->{$subj}->[0]->{$pred}
	&& exists $self->{$subj}->[0]->{$pred}->{$obj}) {
	delete $self->{$subj}->[0]->{$pred}->{$obj};
	delete $self->{$obj}->[1]->{$pred}->{$subj};
    }
}

sub test {
    my($self, $subj,$pred,$obj) = @_;
    if (exists $self->{$subj}
	&& exists $self->{$subj}->[0]->{$pred}
	&& exists $self->{$subj}->[0]->{$pred}->{$obj}) {
	return 1;
    }
    return;
}

sub query3 {
    my($self, $subj,$pred,$obj) = @_;
    my @res;			# List of lists
    if (defined $subj) {
	return unless $self->{$subj};
	if (defined $pred) {
	    my $oh = $self->{$subj}->[0]->{$pred} or return;
	    if (defined $obj) {
		# spo
		push @res, [$subj,$pred,$obj]
		    if return $oh->{$obj};
	    } else {
		# sp?
		@res = map {[$subj,$pred,$_]} keys %$oh;
	    }
	} elsif (defined $obj) {
	    # s?o
	    while (my($p,$oh) = each %{$self->{$subj}->[0]}) {
		push @res, [$subj,$p,$obj] if $oh->{$obj}
	    }
	} else {
	    # s??
	    while (my($p,$oh) = each %{$self->{$subj}->[0]}) {
		foreach my $o (keys %$oh) {
		    push @res, [$subj,$p,$o];
		}
	    }
	}
    } elsif (defined $obj) {
	return unless $self->{$obj};
	if (defined $pred) {
	    # ?po
	    my $sh = $self->{$subj}->[1]->{$pred} or return;
	    @res = map {[$_,$pred,$obj]} keys %$sh;
	} else {
	    # ??o
	    while (my($p,$sh) = each %{$self->{$obj}->[1]}) {
		foreach my $s (keys %$sh) {
		    push @res, [$s,$p,$obj];
		}
	    }
	}
    } elsif (defined $pred) {
	# ?p?
	while (my($s,$sl) = each %$self) {
	    if (exists $sl->[0]->{$pred}) {
		push @res, map {[$s,$pred,$_]} keys %{$sl->[0]->{$pred}};
	    }
	}
    } else {
	# ???
	while (my($s,$sl) = each %$self) {
	    while (my($p,$oh) = each %{$sl->[1]}) {
		foreach my $o (keys %$oh) {
		    push @res, [$s,$p,$o];
		}
	    }
	}
    }
    return wantarray ? @res : $res[0];
}

sub query {
    my($self, $subj,$pred,$obj) = @_;
    my @res;			# List of lists
    my $pat = join("",map {defined($_) || '?'} $subj,$pred,$obj);
    if (defined $subj) {
	return unless $self->{$subj};
	if ($pat eq '111') {
	    return $self->test($subj,$pred,$obj);
	} elsif ($pat eq '11?') {
	    @res = map {[$_]} $self->sp2o($subj,$pred);
	} elsif ($pat eq '1??') {
	    while (my($p,$oh) = each %{$self->{$subj}->[0]}) {
		foreach my $o (keys %$oh) {
		    push @res, [$p,$o];
		}
	    }
	} else {
	    # 1?1
	    while (my($p,$oh) = each %{$self->{$subj}->[0]}) {
		push @res, [$p] if $oh->{$obj}
	    }
	}
    } elsif (defined $obj) {
	return unless $self->{$obj};
	if ($pat eq '?11') {
	    @res = map {[$_]} $self->po2s($pred,$obj);
	} else {
	    # ??1
	    while (my($p,$sh) = each %{$self->{$obj}->[1]}) {
		foreach my $s (keys %$sh) {
		    push @res, [$s,$p,$obj];
		}
	    }
	}
    } elsif (defined $pred) {
	# ?1?
	while (my($s,$sl) = each %$self) {
	    if (exists $sl->[0]->{$pred}) {
		push @res, map {[$s,$_]} keys %{$sl->[0]->{$pred}};
	    }
	}
    } else {
	# ???
	while (my($s,$sl) = each %$self) {
	    while (my($p,$oh) = each %{$sl->[1]}) {
		foreach my $o (keys %$oh) {
		    push @res, [$s,$p,$o];
		}
	    }
	}
    }
    return wantarray ? @res : $res[0];
}

sub sp2o {
    my($self, $subj,$pred) = @_;
    if (exists $self->{$subj} &&
	exists $self->{$subj}->[0]->{$pred}) {
	my @obj = keys %{$self->{$subj}->[0]->{$pred}};
	return wantarray ? @obj : $obj[0];
    }
    return;
}

sub po2s {
    my($self, $subj,$pred) = @_;
    if (exists $self->{$obj} &&
	exists $self->{$obj}->[1]->{$pred}) {
	my @subj = keys %{$self->{$obj}->[1]->{$pred}};
	return wantarray ? @subj : $subj[0];
    }
    return;
}

sub test {
    my($self, $subj,$pred) = @_;
    if (exists $self->{$subj} &&
	exists $self->{$subj}->[0]->{$pred} &&
	exists $self->{$subj}->[0]->{$pred}->{$obj}) {
	return 1;
    }
    return;
}

sub commit {}
sub rollback {
    die ref(shift)." cannot rollback - all updates instantaneous";
}

1;
